class QuiverRepresentation(SageObject):
    def __init__(self, R, V=None):
        r"""

        INPUT:

        - `R` -- the ring over which the representation is defined
        - `V` -- a dictionary pairing an edge with a matrix

        EXAMPLES::

            sage: K = QQbar
            sage: QuiverRepresentation(K, {(0,1) : matrix([1]), matrix([0])})
        """
        # if a ring is given, then change the base ring of all the matrices
        # to the given ring
        if V is None:
            V = R
            R = V.values()[0].base_ring()
        else:
            for (x, M) in V.iteritems():
                V[x] = M.change_ring(R)
        self._base_ring = R

        # construct the representation
        repn = DiGraph()
        for (x,y) in V:
            Vxy = V[x,y]
            repn.add_edge(x, y, Vxy)
            repn.set_vertex(x, R**Vxy.ncols())
            repn.set_vertex(y, R**Vxy.nrows())
        self._repn = repn

    def __getitem__(self, item):
        if isinstance(item, (list,tuple)):
            return self._repn.edge_label(*item)
        else:
            return self._repn.get_vertex(item)

    def __iter__(self):
        for (x,y,m) in self._repn.edges():
            yield (x,y), m

    def quiver(self):
        return self._repn.copy()

    def plot(self, save_pos=True):
        return self._repn.plot(vertex_labels=True, edge_labels=True, save_pos=True)

    def dimension_vector(self):
        return vector(W[i].dimension() for i in range(W._repn.num_verts()))

    def is_source(self, x):
        return self._repn.in_degree(x) == 0

    def is_sink(self, x):
        return self._repn.out_degree(x) == 0

    def reflection_functor(self, x, verbose=False):
        r"""
        Reflect in the vertex x
        """
        if self.is_source(x):
            if verbose is True:
                print "reflecting at the source vertex %s" % x
            return self._reflection_functor_source(x, verbose=verbose)
        elif self.is_sink(x):
            if verbose is True:
                print "reflecting at the sink vertex %s" % x
            return self._reflection_functor_sink(x, verbose=verbose)
        else:
            if verbose is True:
                print "Warning : vertex is neither a source nor a sink (%s)" % x
            return self

    def _reflection_functor_sink(self, x, verbose=False):
        repn = self._repn
        in_edges = repn.incoming_edges(x)

        # change the vertor space at x
        augment = lambda a,b : a.augment(b)
        phi = reduce(augment, [f for (u,v,f) in in_edges])
        ker = phi.right_kernel()
        if verbose:
            s = "\nphi :\n%s \n\nker(phi) :\n%s\n" % (phi, ker)
        repn.set_vertex(x, ker)

        # change the morphisms incident to x
        kerbasis = ker.basis_matrix().transpose()
        d = 0
        for (u,v,f) in in_edges: # note : v == x here
            repn.delete_edge(u,v,f)
            proj = kerbasis[d:d+f.ncols()]
            repn.add_edge(v,u,proj)
            if verbose:
                s += "\nmorphism %s -> %s:\n%s\n" % (v, u, proj)
            d += f.ncols()

        if verbose:
            return s

    def _reflection_functor_source(self, x, verbose=False):
        repn = self._repn
        out_edges = repn.outgoing_edges(x)

        stack = lambda a,b : a.stack(b)
        psi = reduce(stack, [f for (u,v,f) in out_edges])
        img = psi.column_space()
        V = img.ambient_vector_space()
        coker = V.quotient(img)
        if verbose:
            s = "\npsi:\n%s\n\nimage(psi):\n%s\n\ncoker(psi):\n%s\n" % (psi, img, coker)
        repn.set_vertex(x, coker)

        # change the morphisms incident to x
        basismat = coker.quotient_map().matrix()
        d = 0
        for (u,v,f) in out_edges: # note : u == x here
            repn.delete_edge(u,v,f)
            g = basismat[d:d+f.nrows()].transpose()
            repn.add_edge(v,u,g)
            if verbose:
                s += "\nmorphism %s -> %s:\n%s\n" % (v, u, g)
            d += f.nrows()

        if verbose:
            return s

    def interact(self):
        vertices = self._repn.vertices()
        global _first_time
        _first_time = True
        @interact
        def f(n=selector(vertices,nrows=1,default=0, label="apply reflection functor at vertex "), show_calculation=True):
            # HACK to not do anything on the first execution
            global _first_time
            if _first_time is True:
                _first_time = False
                print "dimension vector:", self.dimension_vector()
                show(self.plot())
                return

            orig = self.plot()
            s = self.reflection_functor(n, verbose=True)
            print "\ndimension vector:", self.dimension_vector()
            show(self.plot())
            if show_calculation:
                print "-----------------"
                print "Calculation trace"
                print "-----------------"
                show(orig)
                print s

    interactive=interact

_first_time = True
